unit uFileList;
{$INCLUDE 'FRename3.inc'}

interface

uses
  Windows, SysUtils, Classes, Generics.Collections, Generics.Defaults,
  {$IFDEF FAR3} Plugin3, {$ELSE} PluginW, {$ENDIF} FarKeysW, FarColor,
  uLog, uFAR, uSystem, uFiles, uMessages, uFRenameConsts;

type
  TFileList = class
  private
    fListFileName: string;
    fSourceDirectory: string;
    fEditorID: Integer;
  protected
    function DecodeEditorLine(const Line: string; out SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength: integer): boolean; overload;
    function DecodeEditorLine(const Line: string; out SourceFileName, DestinationFileName: string): boolean; overload;
    function GetFileAttributes(const FileName: string; out Attributes: DWORD): boolean;
  public
    class function FindActiveFileListByEditorID(const AEditorID: integer; out List: TFileList): boolean;
    class function FindActiveFileListByListFileName(const AListFileName: string; out List: TFileList): boolean;
    constructor Create;
    destructor Destroy; override;
    function LoadFromPanel(AActivePanel: boolean): boolean; overload;
    function LoadFromPanel(const AInfo: TFarPanelInfo): boolean; overload;
    function OpenInEditor: boolean;
    function RenameFiles: boolean;
    function SelectDestinationFilenames: boolean;
    function JumpToDestinationFileName: boolean;
    function ShowEditorMenu: boolean;
    property EditorID: integer read fEditorID write fEditorID;
    property ListFileName: string read fListFileName write fListFileName;
    property SourceDirectory: string read fSourceDirectory write fSourceDirectory;
  end;

implementation

type
  TActiveFileLists = class(TObjectList<TFileList>);

var
  ActiveFileLists: TActiveFileLists = nil;

{ TFileList }

constructor TFileList.Create;
begin
  inherited Create;
  ActiveFileLists.Add(Self);
end;

function TFileList.DecodeEditorLine(const Line: string; out SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength: integer): boolean;
var
  i, n: integer;
begin
  Result := False;
  n := Length(Line);
  if n > 0 then
    if Line[1] = '"' then begin
      SourceFileNameStart := 2;
      i := 2;
      while (i <= n) and (Line[i] <> '"') do
        Inc(i);
      SourceFileNameLength := i - SourceFileNameStart;
      Inc(i);
      while (i <= n) and (Line[i] = ' ') do
        Inc(i);
      if (i <= n) and (Line[i] = '"') then begin
        Inc(i);
        DestinationFileNameStart := i;
        while (i <= n) and (Line[i] <> '"') do
          Inc(i);
        DestinationFileNameLength := i - DestinationFileNameStart;
        Inc(i);
        while (i <= n) and (Line[i] = ' ') do
          Inc(i);
        if Pred(i) = n then
          Result := True;
        end;
      end;
end;

function TFileList.DecodeEditorLine(const Line: string; out SourceFileName, DestinationFileName: string): boolean;
var
  SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength: integer;
begin
  Result := False;
  if DecodeEditorLine(Line, SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength) then
    if SourceFileNameLength > 0 then
      if DestinationFileNameLength > 0 then begin
        SourceFileName := Copy(Line, SourceFileNameStart, SourceFileNameLength);
        DestinationFileName := Copy(Line, DestinationFileNameStart, DestinationFileNameLength);
        Result := True;
      end;
end;

destructor TFileList.Destroy;
var
  i: Integer;
begin
  for i := Pred(ActiveFileLists.Count) downto 0 do
    if ActiveFileLists[i] = Self then
      ActiveFileLists.Delete(i);
  if ListFileName <> '' then
    DeleteFile(ListFileName);
  inherited;
end;

class function TFileList.FindActiveFileListByEditorID(const AEditorID: integer; out List: TFileList): boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to Pred(ActiveFileLists.Count) do
    if ActiveFileLists[i].EditorID = AEditorID then begin
      List := ActiveFileLists[i];
      Result := True;
      Break;
    end;
end;

class function TFileList.FindActiveFileListByListFileName(const AListFileName: string; out List: TFileList): boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to Pred(ActiveFileLists.Count) do
    if ActiveFileLists[i].ListFileName = AListFileName then begin
      List := ActiveFileLists[i];
      Result := True;
      Break;
    end;
end;

function TFileList.GetFileAttributes(const FileName: string; out Attributes: DWORD): boolean;
var
  LastError: DWORD;
  SR: TSearchRec;
begin
  Attributes := Windows.GetFileAttributes(PChar(FileName));
  if Attributes <> INVALID_FILE_ATTRIBUTES then
    Result := True
  else begin
    LastError := GetLastError;
    if (LastError = ERROR_FILE_NOT_FOUND) or (LastError = ERROR_PATH_NOT_FOUND) or (LastError = ERROR_INVALID_NAME) then
      Result := False
    else
      if FindFirst(FileName, faAnyFile, SR) = 0 then
        try
          Attributes := SR.Attr;
          Result := True;
        finally
          SysUtils.FindClose(SR);
        end
      else
        Result := False;
  end;
end;

function TFileList.JumpToDestinationFileName: boolean;
var
  EditorInfo: TEditorInfo;
  Line: string;
  SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength: integer;
begin
  Result := False;
  if TFarEditor.GetInfo(EditorID, EditorInfo) then
    if TFarEditor.GetString(EditorID, EditorInfo.CurLine, Line) then
      if DecodeEditorLine(Line, SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength) then
        if TFarEditor.SetCursorPosition(EditorID, EditorInfo.CurLine, DestinationFileNameStart-1) then begin
          TFarEditor.Refresh(EditorID);
          Result := True;
        end;
end;

function TFileList.LoadFromPanel(AActivePanel: boolean): boolean;
var
  Panel: TFarPanelInfo;
begin
  Panel := TFarPanelInfo.Create(AActivePanel);
  try
    Result := LoadFromPanel(Panel);
  finally
    FreeAndNil(Panel);
  end;
end;

function TFileList.LoadFromPanel(const AInfo: TFarPanelInfo): boolean;

  function GetTranslatedFileName(const FileName: string): string;
    begin
      Result := AnsiQuotedStr(FileName, '"');
    end;

  procedure AddFileName(Text: TTextStream; const FileName: string; MaxFileNameLength: integer);
    begin
      Text.WriteLine(Format('%-*.*s    %s', [MaxFileNameLength, MaxFileNameLength, FileName, FileName]));
    end;

var
  Text: TTextStream;
  Description: string;
  i, n, MaxLength: integer;
  SelectedItems: boolean;
begin
  Result := False;
  ListFileName := '';
  MaxLength := 0;
  for i := 0 to Pred(AInfo.Items.Count) do
    if AInfo.Items[i].Selected then begin
      n := Length(GetTranslatedFileName(AInfo.Items[i].FileName));
      if n > MaxLength then
        MaxLength := n;
    end;
  SelectedItems := MaxLength > 0;
  if not SelectedItems then begin
    MaxLength := Length(GetTranslatedFileName(AInfo.Items[AInfo.CurrentItemIndex].FileName));
  end;
  if MaxLength > 0 then begin
    Description := TFarUtils.CopyString(GetMsg(MFileListDescription));
    Description := StringReplace(Description, '${DIRECTORY}', AInfo.Directory, [rfReplaceAll]);
    SourceDirectory := IncludeTrailingPathDelimiter(TFarUtils.FullFileName(AInfo.Directory, True));
    ListFileName := SystemTempFile(SystemTempDir, 'FRename3', '.txt');
    Text := TTextStream.Create(ListFileName, fmCreate);
    try
      Text.Encoding := feUtf8;
      Text.WriteLine(Description);
      if SelectedItems then begin
        for i := 0 to Pred(AInfo.Items.Count) do begin
          if AInfo.Items[i].Selected then begin
            AddFileName(Text, GetTranslatedFileName(AInfo.Items[i].FileName), MaxLength);
          end;
        end;
      end
      else begin
        AddFileName(Text, GetTranslatedFileName(AInfo.Items[AInfo.CurrentItemIndex].FileName), MaxLength);
      end;
      Result := True;
    finally
      FreeAndNil(Text);
    end;
  end;
end;

function TFileList.OpenInEditor: boolean;
{$IFDEF FAR3_UP}
const
  KEY_ALT_F1: TFarKey = (VirtualKeyCode: VK_F1; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
  KEY_ALT_F2: TFarKey = (VirtualKeyCode: VK_F2; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
{$ENDIF}
var
  EditorInfo: TEditorInfo;
  Description: string;
  i, DescriptionLines: integer;
  {$IFDEF FAR3_UP}
  SetKeyBarTitles: TFarSetKeyBarTitles;
  KeyBarTitles: TKeyBarTitles;
  KeyBarLabels: packed array[0..1] of TKeyBarLabel;
  {$ELSE}
  KeyBarTitles: TKeyBarTitles;
  {$ENDIF}
begin
  Result := False;
  if (ListFileName <> '') and FileExists(ListFileName) then begin
    Description := TFarUtils.CopyString(GetMsg(MFileListDescription));
    DescriptionLines := 1;
    for i := 1 to Length(Description) do
      if Description[i] = #10 then
        Inc(DescriptionLines);
    if FarApi.Editor(PFarChar(ListFileName), GetMsg(MTitle), 0, 0, -1, -1,
          EF_NONMODAL or EF_IMMEDIATERETURN or EF_DELETEONLYFILEONCLOSE or EF_DISABLEHISTORY {$IFDEF FAR3_UP} or EF_DISABLESAVEPOS {$ENDIF} ,
          DescriptionLines+1, 0, CP_UTF8) = EEC_MODIFIED
    then begin
      if TFarEditor.GetInfo(-1, EditorInfo) then begin
        Self.fEditorID := EditorInfo.EditorID;
        {$IFDEF FAR3_UP}
        KeyBarLabels[0].Key := KEY_ALT_F1;
        KeyBarLabels[0].Text := GetMsg(MKeyJumpToDestinationFileName);
        KeyBarLabels[0].LongText := GetMsg(MKeyJumpToDestinationFileNameLong);
        KeyBarLabels[1].Key := KEY_ALT_F2;
        KeyBarLabels[1].Text := GetMsg(MKeySelectDestinationFileNames);
        KeyBarLabels[1].LongText := GetMsg(MKeySelectDestinationFileNamesLong);
        KeyBarTitles.CountLabels := Length(KeyBarLabels);
        KeyBarTitles.Labels := @KeyBarLabels[0];
        SetKeyBarTitles.StructSize := Sizeof(SetKeyBarTitles);
        SetKeyBarTitles.Titles := @KeyBarTitles;
        FarApi.EditorControl(EditorInfo.EditorID, ECTL_SETKEYBAR, 0, @SetKeyBarTitles);
        {$ELSE}
        FillChar(KeyBarTitles, Sizeof(KeyBarTitles), 0);
        KeyBarTitles.AltTitles[0] := GetMsg(MKeyJumpToDestinationFileName);
        KeyBarTitles.AltTitles[1] := GetMsg(MKeySelectDestinationFileNames);
        FarApi.EditorControl(ECTL_SETKEYBAR, @KeyBarTitles);
        {$ENDIF}
        Result := True;
      end
    else
      FarMessage([MTitle, MFailedToLocateEditor, MOK], FMSG_WARNING, 1);
    end
    else
      FarMessage([MTitle, MFailedToOpenEditor, MOK], FMSG_WARNING, 1);
  end;
end;

function TFileList.RenameFiles: boolean;
var
  Text: TTextStream;
  Line, SourceFileName, DestinationFileName, SrcFN, DestFN: string;
  SrcAttr, DestAttr: DWORD;
  Again, AlwaysSkip: boolean;
begin
  Result := False;
  if (ListFileName <> '') and FileExists(ListFileName) then begin
    Text := TTextStream.Create(ListFileName, fmOpenRead or fmShareDenyNone);
    try
      AlwaysSkip := False;
      Result := True;
      while Text.ReadLine(Line) do
        if DecodeEditorLine(Line, SourceFileName, DestinationFileName) then
          repeat
            Again := False;
            SrcFN := SourceDirectory + SourceFileName;
            DestFN := SourceDirectory + DestinationFileName;
            if SrcFN = DestFN then
              Break
            else if not GetFileAttributes(SrcFN, SrcAttr) then
              TFarUtils.ShowMessage([GetMsg(MTitle), GetMsg(MSourceFileNotFound), PFarChar(SrcFN), GetMsg(MOk)], FMSG_WARNING, 1)
            else begin
              if GetFileAttributes(DestFN, DestAttr) and ((DestAttr and FILE_ATTRIBUTE_DIRECTORY) <> 0) and (AnsiCompareText(SrcFN, DestFN) <> 0) then begin
                TFarUtils.ShowMessage([GetMsg(MTitle), GetMsg(MCantOverwriteDirectory), PFarChar(DestFN), GetMsg(MOk)], FMSG_WARNING, 1);
                Result := False;
              end
              else if not RenameFile(SrcFN, DestFN) then
                if AlwaysSkip then
                  Result := False
                else
                  case TFarUtils.ShowMessage([GetMsg(MTitle), GetMsg(MFailedToRename), PFarChar(SrcFN), GetMsg(MFailedToRenameTo), PFarChar(DestFN), GetMsg(MRetry), GetMsg(MOverwrite), GetMsg(MSkip), GetMsg(MSkipAll)], FMSG_WARNING, 4) of
                    0:   Again := True;
                    1:   begin
                           DeleteFile(DestFN);
                           Again := True;
                         end;
                    2:   Result := False;
                    3:   begin
                           AlwaysSkip := True;
                           Result := False;
                         end;
                    else Result := False;
                  end;
            end;
          until not Again;
    finally
      FreeAndNil(Text);
    end;
  end;
end;

function TFileList.SelectDestinationFilenames: boolean;
var
  EditorInfo: TEditorInfo;
  Select: TEditorSelect;
  LineNr, SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength: integer;
  Line: string;
begin
  Result := False;
  if TFarEditor.GetInfo(EditorID, EditorInfo) then begin
    FillChar(Select, Sizeof(Select), 0);
    {$IFDEF FAR3_UP}
    Select.StructSize := Sizeof(Select);
    {$ENDIF}
    Select.BlockType := BTYPE_COLUMN;
    Select.BlockStartLine := -1;
    Select.BlockStartPos := -1;
    Select.BlockWidth := -1;
    Select.BlockHeight := -1;
    for LineNr := 0 to Pred(EditorInfo.TotalLines) do
      if TFarEditor.GetString(EditorID, LineNr, Line) then
        if DecodeEditorLine(Line, SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength) then begin
          if Select.BlockStartLine < 0 then
            Select.BlockStartLine := LineNr;
          if (Select.BlockStartPos < 0) or (Select.BlockStartPos > (DestinationFileNameStart-2)) then
            Select.BlockStartPos := DestinationFileNameStart-2;
          if Select.BlockWidth < (DestinationFileNameLength+2) then
            Select.BlockWidth := DestinationFileNameLength+2;
          Select.BlockHeight := LineNr - Select.BlockStartLine + 1;
          Result := True;
        end;
    if Result then begin
      Result := FarApi.EditorControl( {$IFDEF FAR3_UP} EditorID, {$ENDIF} ECTL_SELECT, {$IFDEF FAR3_UP} 0, {$ENDIF} @Select) <> 0;
      TFarEditor.Refresh(EditorID);
    end;
  end;
end;

function TFileList.ShowEditorMenu: boolean;

  procedure SetMenuItem(var MenuItem: TFarMenuItem; TextID: TMessages {$IFDEF FAR3_UP} ; Flags: TMenuItemFlags {$ENDIF} );
    begin
      FillChar(MenuItem, Sizeof(MenuItem), 0);
      MenuItem.TextPtr := GetMsg(TextID);
      {$IFDEF FAR3_UP}
      MenuItem.Flags := Flags;
      {$ELSE}
      MenuItem.Selected := 0;
      MenuItem.Checked := 0;
      MenuItem.Separator := 0;
      {$ENDIF}
    end;

var
  MenuItems: packed array[0..1] of TFarMenuItem;
begin
  SetMenuItem(MenuItems[0], MKeyJumpToDestinationFileNameLong {$IFDEF FAR3_UP} , MIF_NONE {$ENDIF} );
  SetMenuItem(MenuItems[1], MKeySelectDestinationFileNamesLong {$IFDEF FAR3_UP} , MIF_NONE {$ENDIF} );
  case FarApi.Menu( {$IFDEF FAR3_UP} PLUGIN_GUID, MENU_EDITOR_GUID {$ELSE} FarApi.ModuleNumber {$ENDIF} , -1, -1, 0, FMENU_WRAPMODE, GetMsg(MTitle), nil, nil, nil, nil, @MenuItems[0], Length(MenuItems)) of
    0:   Result := JumpToDestinationFileName;
    1:   Result := SelectDestinationFilenames;
    else Result := False;
  end;
end;

initialization
  ActiveFileLists := TActiveFileLists.Create(False);

finalization
  FreeAndNil(ActiveFileLists);

end.

